from os.path import join
from typing import Union, List, Tuple
from pathlib import Path

import pandas as pd
from netCDF4 import Dataset
import xarray as xr
from scipy.ndimage import binary_dilation
import numpy as np

from .utils import file_exists
from .chem import Chemistry
from .array import GriddedData


class GenieModel(object):
    """
    GenieModel is the interface to cGENIE output
    """

    def __init__(self, model_path: Union[str, List, Tuple], gemflag=None):
        """
        Initialise a GenieModel object with a path to cGENIE output directory

        -------
        Example
        >>> from cgeniepy.model import GenieModel
        >>> model = GenieModel("path_to_GENIE_output")

        ## specify cgenie component
        >>> model = GenieModel("path_to_GENIE_output", gemflag=["biogem"])
        """

        ## check if model_path is a valid directory
        if isinstance(model_path, (list, tuple)):
            self.is_ensemble = True
            for path in model_path:
                if not Path(path).is_dir():
                    raise ValueError(f"{path} is not a valid directory")
        elif isinstance(model_path, str):
            self.is_ensemble = False
            if not Path(model_path).is_dir():
                raise ValueError(f"{model_path} is not a valid directory")

        if not gemflag:
            print(">>> No gemflag is provided, use default gemflags: [biogem, ecogem]")
            self.gemflag = ["biogem", "ecogem"]
        else:
            self.gemflag = gemflag
            ## if gemflag is a string, convert it to a list
            if isinstance(gemflag, str):
                self.gemflag = [gemflag]

        self.model_path = model_path

    def _model_ncpath(self, gem="ecogem", dim="2d"):
        """
        get the path of model NetCDF output of target gem and dimension
        if the model is an ensemble, return a list of paths
        """

        nc_file = f"fields_{gem}_{dim}.nc"

        if not self.is_ensemble:
            model_path = self.model_path
            nc_path = join(model_path, gem, nc_file)
            if file_exists(nc_path):
                return nc_path
        else:
            nc_paths = []
            for path in self.model_path:
                model_path = path
                nc_path = join(model_path, gem, nc_file)
                if file_exists(nc_path):
                    nc_paths.append(nc_path)
            ## make `nc_paths` hashable
            nc_paths = tuple(nc_paths)
            return nc_paths

    def ncvar_dict(self) -> dict:
        """
        Return all available variables and related biogem/ecogem (2d or 3d)
        NetCDF path for each model
        """

        ## initialise a dictionary, key: model_path, value: ncvar_list
        var_path = {}

        for gem in self.gemflag:
            for dim in ["2d", "3d"]:
                ## ignore ecogem 3d which is not available now (Oct 2023)
                if gem == "ecogem" and dim == "3d":
                    continue

                if self.is_ensemble:
                    ## use one model to get the nc_path
                    nc_path = self._model_ncpath(gem, dim)[0]
                    all_nc_path = self._model_ncpath(gem, dim)
                else:
                    nc_path = self._model_ncpath(gem, dim)

                nc = Dataset(nc_path, "r")
                ncvar_list = list(nc.variables.keys())
                nc.close()

                ## add key, value to the dictionary
                if self.is_ensemble:
                    var_path[all_nc_path] = ncvar_list
                else:
                    var_path[nc_path] = ncvar_list

        return var_path

    def _lookup_ncpath(self, var):
        """
        find the netcdf path of the target variable

        if more than one model satisfies the condition,
        return all model's NetCDF paths
        """
        for path, value_list in self.ncvar_dict().items():
            if var in value_list:
                return path
        raise ValueError(f"variable {var} not found in the ncvar_dict")

    def _open_nc(self, nc_path):
        "Use xarray to open netcdf file"
        ## if path is a list of paths
        if not isinstance(nc_path, (list, tuple)):
            ## `open_dataset` to lazy load the data
            return xr.open_dataset(nc_path)
        else:
            datasets = [xr.open_dataset(file) for file in nc_path]
            combined_ds = xr.concat(datasets, "model")
            return combined_ds

    def get_var(self, var: Union[str, List, Tuple], unit=None):
        """
        Get the data of target variable.
        A list of variables is supported as well.

        :param var: the name of target variable
        :param unit: the unit of target variable, usually provided in the model output
        """
        self.target_var = var

        ## if varstr is a string
        if isinstance(self.target_var, str):
            ## find the path to the netcdf file
            path2nc = self._lookup_ncpath(var=self.target_var)
            ## open the netcdf file
            array = self._open_nc(path2nc)[self.target_var]
        ## if varstr is a list/tuple of strings
        ## concat the data along the "variable" dimension
        elif isinstance(self.target_var, (list, tuple)):
            array_container = []
            for v in self.target_var:
                path2nc = self._lookup_ncpath(var=v)
                single_array = self._open_nc(path2nc)[v]
                array_container.append(single_array)

            array = xr.concat(array_container, "variable")
            array.name = "ensemble_variable"

        ## initialise GenieArray
        target_data = GriddedData(array)
        try:
            c = Chemistry()
            target_data.array.attrs["units"] = c.format_unit(
                target_data.array.attrs["units"]
            )
        except KeyError:
            print("Unit not found in cGENIE output, please check the FORTRAN code!")

        ## add unit if provided
        if not unit:
            target_data.array.attrs["units"] = unit

        return target_data

    def tsvar_list(self):
        """list all files biogem timeseries files

        If the model is an ensemble, it assumes that all models share the same
        file/data structure and only return the first model's timeseries list
        """

        if self.is_ensemble:
            biogem_path = join(self.model_path[0], "biogem")
        else:
            biogem_path = join(self.model_path, "biogem")

        ## return all files in the directory
        all_bg_files = Path(biogem_path)

        ## find out files with .res extension
        tsvar_list = [f for f in all_bg_files.glob("*.res")]

        return tsvar_list

    def get_ts(self, filename: str):
        """
        read in time series output of GENIE

        :param filename: the name of the time series file
        :return: a pandas DataFrame
        """

        if not self.is_ensemble:
            f = join(self.model_path, "biogem", filename)
            if not file_exists(f):
                raise ValueError(f"{f} does not exist")

            # Read the text file into a DataFrame
            with open(f, "r") as file:
                lines = file.readlines()

            # Remove " %" from the header and split it using " / "
            # not '/' which is also included in the isotope unit (e.g. 'd13C o/oo')
            header = [col.strip().lstrip("% ") for col in lines[0].split(" / ")]
            data = [line.split() for line in lines[1:]]
            df = pd.DataFrame(data, columns=header)

            ## convert to numeric
            df = df.apply(pd.to_numeric, errors="coerce")

            ## add a column of model name
            df["model"] = self.model_path.split("/")[-1]
            return df
        else:
            ## concatenate all models' data into one data frame
            df_list = []

            for path in self.model_path:
                ## do the same thing as above
                f = join(path, "biogem", filename)
                if not file_exists(f):
                    raise ValueError(f"{f} does not exist")

                ## read text file into a DataFrame
                with open(f, "r") as file:
                    lines = file.readlines()

                ## Remove " %" from the header and split it using " / "
                header = [col.strip().lstrip("%") for col in lines[0].split(" / ")]
                data = [line.split() for line in lines[1:]]

                df = pd.DataFrame(data, columns=header)

                ## convert to numeric
                df = df.apply(pd.to_numeric, errors="coerce")

                df["model"] = path.split("/")[-1]
                df_list.append(df)

            ## concatenate by row
            all_df = pd.concat(df_list, axis=0)

            return all_df

    def grid_mask(self):
        """
        cGENIE continent mask array (by setting zero values),
        either calculated from existing data or using biogem.grid_mask
        """
        try:
            grid_mask = self.get_var("grid_mask").array
            return grid_mask
        except ValueError:
            print("grid_mask not found!")

    def grid_category(self):
        """an alogirthm to define surface grid catogories depending on the land-sea mask
        0: coastal sea
        1: land
        2: open ocean
        """

        is_sea = np.isnan(self.grid_mask())
        land_around = binary_dilation(is_sea)
        grid_catogories = xr.where(land_around, np.logical_and(land_around, is_sea), 2)

        return grid_catogories

    def grid_area(self):
        "grid area array in m2"
        ## read grid_area from biogem
        print("grid area returned in the unit of 'm2'")
        return self.get_var("grid_area")

    def grid_volume(self):
        "grid volume array (3d) in m3"
        grid_volume = self.grid_area()  ## m2
        depth = self.get_var("zt")  ## m
        try:
            grid_volume = grid_volume * depth
            grid_volume.array.attrs["units"] = "m$^{3}$"
            grid_volume.array.attrs["long_name"] = "grid volume"
            print("grid volume calculated in the unit of 'm3'")
        except ValueError:
            print(
                "Depth array not found! Please ensure 3d data is exported in the model!"
            )

        return grid_volume



import pathlib
import xml.etree.ElementTree as ET

import numpy as np
import pandas as pd
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import xarray as xr

import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import ListedColormap, to_rgb as hex_to_rgb
import mpl_toolkits.axisartist.floating_axes as fa
import mpl_toolkits.axisartist.grid_finder as gf

from .utils import efficient_log
from .chem import Chemistry

class ArrayVis:

    transform_crs = ccrs.PlateCarree()  # do not change

    def __init__(self, array):

        self.array = array

        if hasattr(self.array, "units"):
            self.units = Chemistry().format_unit(self.array.units)
        else:
            self.units = None

        if hasattr(self.array, "long_name"):
            self.long_name = self.array.long_name
        else:
            self.long_name = None

        ## aesthetic parameters
        self.aes_dict = {
            ## x,ylabel
            "general_kwargs": {"cmap": "viridis", "font": "Helvetica", "fontsize": 10},
            "facecolor_kwargs": {"c": "silver"},
            "borderline_kwargs": {"c": "black", "linewidth": 0.5},
            "outline_kwargs": {"colors": "black", "linewidth": 0.5},
            "gridline_kwargs": {"colors": "gray", "linewidth": 0.5},
            "pcolormesh_kwargs": {"shading": "auto", "cmap": plt.get_cmap("viridis")},
            "contour_kwargs": {
                "linewidths": 0.6,
                "colors": "black",
                "linestyles": "solid",
                "zorder": 10,
            },
            "contour_label_kwargs": {
                "colors": ["black"],
                "fontsize": 8,
                "inline": False,
            },
            "contourf_kwargs": {"levels": 20},
            "colorbar_label_kwargs": {
                "label": f"{self.long_name}\n({self.units})",
                "size": 10,
                "labelpad": 10,
            },
            "colorbar_kwargs": {"fraction": 0.046, "pad": 0.04},
        }

    def plot(self, *args, **kwargs):
        """
        visualise the data based on the dimension of the array

        *args and **kwargs are passed to plotting functions
        in practical, turn on/off the plotting elements
        e.g., plot(x, y, pcolormesh=False, contour=True)
        each plotting function will seek for the corresponding kwargs
        in self.aes_dict
        """

        if self.array.ndim == 1:
            return self._plot_1d(*args, **kwargs)
        elif self.array.ndim == 2:
            return self._plot_2d(*args, **kwargs)
        elif self.array.ndim == 3:
            return self._plot_3d(*args, **kwargs)
        else:
            raise ValueError(f"{self.ndim} dimensions not supported")

    def _plot_1d(self, *args, **kwargs):
        """
        plot 1D data, e.g., zonal_average, time series
        dim: time/lon/lat
        """
        if "ax" not in kwargs:
            fig, local_ax = self._init_fig()
        else:
            local_ax = kwargs.pop("ax")

        ## get the only dimension as x
        dim = self.array[self.array.dims[0]]

        if not "swap_xy" in kwargs:
            local_ax.set_xlabel(dim.name)
            local_ax.set_ylabel(f"{self.array.long_name} ({self.array.units})")
            p = local_ax.plot(dim, self.array, *args, **kwargs)
        else:
            local_ax.set_xlabel(f"{self.array.long_name} ({self.array.units})")
            local_ax.set_ylabel(dim.name)
            p = local_ax.plot(self.array, dim, *args, **kwargs)

        return p

    def _plot_2d(self, *arg, **kwargs):
        "plot 2D data, e.g., map, trasect"
        ## if lon, lat then plot map
        ## if lon, zt then plot transect
        ## if lat, zt then plot transect
        dims = self.array.dims
        if "lon" in dims and "lat" in dims:
            return self._plot_map(*arg, **kwargs)
        elif "zt" in dims and "lon" in dims:
            return self._plot_transect(*arg, **kwargs)
        elif "zt" in dims and "lat" in dims:
            return self._plot_transect(*arg, **kwargs)
        else:
            raise ValueError(f"{dims} not supported")

    def _plot_3d(self, *args, **kwargs):
        "plot 3D data = plot mutiple 2D plots"
        print("3D plot not supported yet")
        pass

    def _plot_map(
        self,
        pcolormesh=True,
        contour=False,
        colorbar=False,
        contourf=False,
        outline=False,
        facecolor=True,
        borderline=True,
        gridline=False,
        *args,
        **kwargs,
    ):

        x_name = self.array.dims[1]  ## lon
        y_name = self.array.dims[0]  ## lat

        x_arr = self.array[x_name]
        y_arr = self.array[y_name]

        x_edge = np.linspace(-260, 100, x_arr.size + 1)
        y_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, y_arr.size + 1)))

        if "ax" not in kwargs:
            fig, local_ax = self._init_fig(subplot_kw={"projection": ccrs.EckertIV()})
        else:
            local_ax = kwargs.pop("ax")

        if facecolor:
            self._set_facecolor(local_ax, **self.aes_dict["facecolor_kwargs"])

        if borderline:
            self._set_borderline(
                local_ax, geo=True, **self.aes_dict["borderline_kwargs"]
            )

        if outline:
            ## outline uses edge coordinates
            ## need to be transformed to PlateCarree
            self._add_outline(
                local_ax,
                x=x_edge,
                y=y_edge,
                transform=self.transform_crs,
                **self.aes_dict["outline_kwargs"],
            )

        if gridline:
            self._add_gridline(
                local_ax,
                transform=self.transform_crs,
                **self.aes_dict["gridline_kwargs"],
            )

        if pcolormesh:
            ## pcolormesh uses edge coordinates
            ## need to be transformed to PlateCarree
            p_pcolormesh = self._add_pcolormesh(
                local_ax,
                x=x_edge,
                y=y_edge,
                transform=self.transform_crs,
                *args,
                **self.aes_dict["pcolormesh_kwargs"],
            )
            if colorbar:
                cbar = self._add_colorbar(p_pcolormesh, orientation="horizontal")
                self._add_colorbar_label(cbar, **self.aes_dict["colorbar_label_kwargs"])

        if contour:
            ## contour uses center coordinates
            ## need to be transformed to PlateCarree
            p_contour = self._add_contour(
                local_ax,
                x_arr,
                y_arr,
                transform=self.transform_crs,
                **self.aes_dict["contour_kwargs"],
            )
            self._add_contour_label(
                local_ax, p_contour, **self.aes_dict["contour_label_kwargs"]
            )
            ## contour will not be used to plot colorbar because it's set to black

        if contourf:
            p_contourf = self._add_contourf(
                local_ax,
                x_arr,
                y_arr,
                transform=self.transform_crs,
                **self.aes_dict["contourf_kwargs"],
            )

        return local_ax

    def _plot_transect(
        self,
        x="lat_edge",
        y="zt_edge",
        pcolormesh=True,
        contour=False,
        colorbar=True,
        contourf=False,
        outline=False,
        facecolor=True,
        borderline=True,
        *args,
        **kwargs,
    ):
        """
        Examples

        import matplotlib.pyplot as plt

        model = EcoModel("path_to_model")

        fig, axs=plt.subplots(nrows=1, ncols=3, figsize=(15, 3))

        basins = ['Atlantic', 'Pacific', 'Indian']

        for i in range(3):
            model.get_var('ocn_PO4').isel(time=-1).mask_basin(base='worjh2',basin=basins[i], subbasin='').mean(dim='lon').plot(ax=axs[i])
            axs[i].title.set_text(basins[i])


        available kwargs:
        outline_kwargs = {'outline_color': 'red', 'outline_width': .5}
        """

        x_name = self.array.dims[1]  ## lat
        y_name = self.array.dims[0]  ## zt

        x_arr = self.array[x_name]
        y_arr = self.array[y_name]

        if "ax" not in kwargs:
            fig, local_ax = self._init_fig(figsize=(6, 3))
        else:
            local_ax = kwargs.pop("ax")

        if facecolor:
            self._set_facecolor(local_ax, **self.aes_dict["facecolor_kwargs"])

        if borderline:
            self._set_borderline(
                local_ax, geo=False, **self.aes_dict["borderline_kwargs"]
            )

        if outline:
            ## outline uses edge coordinates
            self._add_outline(
                local_ax, x=x_arr, y=y_arr, **self.aes_dict["outline_kwargs"]
            )

        if pcolormesh:
            ## pcolormesh uses edge coordinates
            p_pcolormesh = self._add_pcolormesh(
                local_ax, x=x_arr, y=y_arr, *args, **self.aes_dict["pcolormesh_kwargs"]
            )
            if colorbar:
                cbar = self._add_colorbar(p_pcolormesh, orientation="vertical")
                self._add_colorbar_label(cbar, **self.aes_dict["colorbar_label_kwargs"])

        if contour:
            ## contour uses center coordinates
            ## need to be transformed to PlateCarree
            p_contour = self._add_contour(
                local_ax, x_arr, y_arr, **self.aes_dict["contour_kwargs"]
            )
            self._add_contour_label(
                local_ax, p_contour, **self.aes_dict["contour_label_kwargs"]
            )
            ## contour will not be used to plot colorbar because it's set to black

        if contourf:
            p_contourf = self._add_contourf(local_ax, x_arr, y_arr)

        ## reverse y axis
        local_ax.set_ylim(local_ax.get_ylim()[::-1])
        local_ax.set_ylabel(
            "Depth (m)",
            fontsize=self.aes_dict["general_kwargs"]["fontsize"],
            font=self.aes_dict["general_kwargs"]["font"],
        )
        local_ax.set_xlabel(
            x.split("_")[0],
            fontsize=self.aes_dict["general_kwargs"]["fontsize"],
            font=self.aes_dict["general_kwargs"]["font"],
        )

        ## x/y tick label
        local_ax.tick_params(
            axis="both",
            which="major",
            labelsize=self.aes_dict["general_kwargs"]["fontsize"],
            labelfontfamily=self.aes_dict["general_kwargs"]["font"],
        )

        return local_ax

    ## ------- Below is implementations -------------------------

    def _init_fig(self, *args, **kwargs):
        return plt.subplots(dpi=120, *args, **kwargs)

    def _add_pcolormesh(self, ax, x, y, *args, **kwargs):
        return ax.pcolormesh(x, y, self.array, *args, **kwargs)

    def _add_contour(self, ax, x, y, *args, **kwargs):
        return ax.contour(x, y, self.array, *args, **kwargs)

    def _add_contourf(self, ax, x, y, *args, **kwargs):
        return ax.contourf(x, y, self.array, *args, **kwargs)

    def _add_contour_label(self, ax, cs, *args, **kwargs):
        ax.clabel(cs, cs.levels[::3], *args, **kwargs)

    def _add_gridline(self, ax, *args, **kwargs):
        ax.gridlines(*args, **kwargs)

    def _set_borderline(self, ax, geo=True, **kwargs):
        if geo:
            ax.spines["geo"].set_edgecolor(kwargs["c"])
            ax.spines["geo"].set_linewidth(kwargs["linewidth"])
        else:
            for direction in ["top", "bottom", "left", "right"]:
                ax.spines[direction].set_edgecolor(kwargs["c"])
                ax.spines[direction].set_linewidth(kwargs["linewidth"])
                # vertical layer order
                ax.spines[direction].set_zorder(0)

    def _set_facecolor(self, ax, **kwargs):
        ax.patch.set_color(**kwargs)

    def _add_outline(self, ax, x, y, **kwargs):
        """
        draw outlines for the NA grids
        """
        mask_array = np.where(~np.isnan(self.array), 1, 0)

        Ny_edge = len(y)
        Nx_edge = len(x)
        # dimension of array
        Ny_array = Ny_edge - 1
        Nx_array = Nx_edge - 1
        # index of array
        Ny_index = Ny_array - 1
        Nx_index = Nx_array - 1

        # i stands for yitude index, j stands for xgitude index
        # e.g., mask_array[0, ] is Antarctic ice cap
        # i,j are for mask array and will be converted to x/y edge lines
        for i in range(Ny_array):
            for j in range(Nx_array):
                # compare with the right grid, and plot vertical line if different
                if j < Nx_index and mask_array[i, j] != mask_array[i, j + 1]:
                    ax.vlines(x[j + 1], y[i], y[i + 1], **kwargs)

                # connect the circular xgitude axis
                if j == Nx_index and mask_array[i, j] != mask_array[i, 0]:
                    ax.vlines(x[j + 1], y[i], y[i + 1], **kwargs)

                # compare with the above grid, and plot horizontal line if different
                if i < Ny_index and mask_array[i, j] != mask_array[i + 1, j]:
                    ax.hlines(y[i + 1], x[j], x[j + 1], **kwargs)

    def _add_colorbar(self, mappable_object, *args, **kwargs):
        cbar = plt.colorbar(mappable_object, *args, **kwargs)
        return cbar

    def _add_colorbar_label(self, cbar, *args, **kwargs):
        ## set colorbar label
        cbar.set_label(*args, **kwargs)
        cbar.ax.tick_params(
            axis="both",
            which="major",
            labelsize=self.aes_dict["general_kwargs"]["fontsize"],
            labelfontfamily=self.aes_dict["general_kwargs"]["font"],
        )
        cbar.outline.set_edgecolor("black")
        cbar.outline.set_linewidth(0.5)


class ScatterVis:
    "Visualisation object based on dataframe"

    def __init__(self, df):
        self.df = df

    def _init_fig(self, *args, **kwargs):
        return plt.subplots(dpi=120, *args, **kwargs)

    def plot_map(
        self,
        ax=None,
        log=False,
        land_mask=True,
        *args,
        **kwargs,
    ):
        """plot map based on dataframe with latitude/longitude
        using cartopy as engine

        :param df: pandas dataframe
        :param var: variable (column) in dataframe to plot
        :param x: coordinate attribute, default "Longitude"
        :param y: coordinate attribute, default "Latitude"
        :param interpolate: whether interpolate scatter data

        :returns: a map
        """
        if not ax:
            fig, ax = self._init_fig(subplot_kw={"projection": ccrs.EckertIV()})

        if land_mask:
            ax.set_global()
            # plot land and coastline, zorder is the drawing order, smaller -> backer layer
            ax.add_feature(
                cfeature.LAND.with_scale("110m"), zorder=2, facecolor="#B1B2B4"
            )
            ax.add_feature(cfeature.COASTLINE.with_scale("110m"), zorder=3)

        if log:
            self.df[self.var] = efficient_log(self.df[self.var])

        if self.df[self.var].dtype != float:
            self.df[self.var] = self.df[self.var].astype(float)

        p = ax.scatter(
            x=self.df[self.lon],
            y=self.df[self.lat],
            c=self.df[self.var],
            transform=ccrs.PlateCarree(),
            *args,
            **kwargs,
        )

        return p

    def plot_transect(
        self,
        ax=None,
        bathy_lon=None,
        *args,
        **kwargs,
    ):
        if not ax:
            fig, ax = self._init_fig()

        if self.df[self.var].dtype != float:
            self.df[self.var] = self.df[self.var].astype(float)
            
        p = ax.scatter(
            x=self.df[self.lat],
            y=self.df[self.depth],
            c=self.df[self.var],
            *args,
            **kwargs,
        )

        

        if bathy_lon:
            data_dir = pathlib.Path(__file__).parent.parent
            bathy = xr.open_dataset(data_dir / "data/GEBCO2002_bathy.nc")

            ## get the lat range
            min_lat = self.df[self.lat].min()
            max_lat = self.df[self.lat].max()
            min_depth = self.df[self.depth].min()

            up = bathy.z.sel(lon=slice(*bathy_lon)).mean(dim="lon")
            # up = up.sel(lat=slice(min_lat, max_lat))
            bottom = np.ones(len(up)) * -5500

            ax.fill_between(up.lat, up, bottom, color="black")

        return p


class CommunityPalette:

    def get_palette(self, cmap_name, N=256, reverse=False, alpha=None):
        """
        community-driven colormaps with multiple sources

        :param cmap_name: colormap name, can be found in avail_palette()
        :type cmap_name: str

        :returns: colormap

        XML data: https://sciviscolor.org/colormaps/
        txt data: from original packages
        """

        if cmap_name not in self.avail_palettes():
            raise ValueError(
                f"{cmap_name} not found, accepted values are {self.avail_palette()}"
            )

        data_dir = pathlib.Path(__file__).parent.parent

        file_path = None
        file_ext = None
        colors = []
        c = None

        # Search for the directory based on list comprehension
        file_path = [f for f in data_dir.glob("**/*") if cmap_name in str(f)]
        if len(file_path) > 1:
            raise ValueError("Multiple colormaps found")
        else:
            file_path = file_path[0]
            file_ext = file_path.suffix

        if file_ext == ".txt":
            colors = pd.read_csv(file_path, header=None).values.tolist()
            colors = [colors[i][0] for i in range(len(colors))]
            c = ListedColormap(colors, name=cmap_name)
        elif file_ext == ".xml":
            tree = ET.parse(file_path)
            root = tree.getroot()
            positions = []

            # Extract color points and their positions from the XML
            for point in root.findall(".//Point"):
                r = float(point.get("r"))
                g = float(point.get("g"))
                b = float(point.get("b"))
                colors.append((r, g, b))
                positions.append(float(point.get("x")))

            c = ListedColormap(colors, name=cmap_name)

        # Optionally add transparency
        if alpha is not None:
            if file_ext == ".txt":
                rgb_array = np.array([hex_to_rgb(i) for i in colors])
                rgb_array[:, -1] = alpha
                c = ListedColormap(rgb_array, N=N)
            elif file_ext == ".xml":
                # Assuming alpha is to be applied uniformly to all colors in the XML-based colormap
                rgba_colors = [
                    (color[0], color[1], color[2], alpha) for color in colors
                ]
                c = ListedColormap(rgba_colors, name=cmap_name)

        if reverse and c is not None:
            return c.reversed()

        ## if _r exsit in the string, reverse the colormap
        if "_r" in cmap_name and c is not None:
            return c.reversed()

        if c is None:
            raise ValueError("Colormap could not be created")

        return c

    def avail_palettes(self):
        """return a list of colormap names"""
        data_dir = pathlib.Path(__file__).parent.parent

        return [
            f.stem
            for f in data_dir.glob("data/colormaps/*")
            if f.suffix in [".txt", ".xml"]
        ]



from re import findall
class Chemistry:

    def molar_mass(self, element):
        "get molar mass (g/mol) of chemistry elements"
        d = {
            "H": 1.00794,
            "He": 4.002602,
            "Li": 6.941,
            "Be": 9.012182,
            "B": 10.811,
            "C": 12.0107,
            "N": 14.0067,
            "O": 15.9994,
            "F": 18.9984032,
            "Ne": 20.1797,
            "Na": 22.98976928,
            "Mg": 24.305,
            "Al": 26.9815386,
            "Si": 28.0855,
            "P": 30.973762,
            "S": 32.065,
            "Cl": 35.453,
            "Ar": 39.948,
            "K": 39.0983,
            "Ca": 40.078,
            "Sc": 44.955912,
            "Ti": 47.867,
            "V": 50.9415,
            "Cr": 51.9961,
            "Mn": 54.938045,
            "Fe": 55.845,
            "Co": 58.933195,
            "Ni": 58.6934,
            "Cu": 63.546,
            "Zn": 65.409,
            "Ga": 69.723,
            "Ge": 72.64,
            "As": 74.9216,
            "Se": 78.96,
            "Br": 79.904,
            "Kr": 83.798,
            "Rb": 85.4678,
            "Sr": 87.62,
            "Y": 88.90585,
            "Zr": 91.224,
            "Nb": 92.90638,
            "Mo": 95.94,
            "Tc": 98.9063,
            "Ru": 101.07,
            "Rh": 102.9055,
            "Pd": 106.42,
            "Ag": 107.8682,
            "Cd": 112.411,
            "In": 114.818,
            "Sn": 118.71,
            "Sb": 121.760,
            "Te": 127.6,
            "I": 126.90447,
            "Xe": 131.293,
            "Cs": 132.9054519,
            "Ba": 137.327,
            "La": 138.90547,
            "Ce": 140.116,
            "Pr": 140.90465,
            "Nd": 144.242,
            "Pm": 146.9151,
            "Sm": 150.36,
            "Eu": 151.964,
            "Gd": 157.25,
            "Tb": 158.92535,
            "Dy": 162.5,
            "Ho": 164.93032,
            "Er": 167.259,
            "Tm": 168.93421,
            "Yb": 173.04,
            "Lu": 174.967,
            "Hf": 178.49,
            "Ta": 180.9479,
            "W": 183.84,
            "Re": 186.207,
            "Os": 190.23,
            "Ir": 192.217,
            "Pt": 195.084,
            "Au": 196.966569,
            "Hg": 200.59,
            "Tl": 204.3833,
            "Pb": 207.2,
            "Bi": 208.9804,
            "Po": 208.9824,
            "At": 209.9871,
            "Rn": 222.0176,
            "Fr": 223.0197,
            "Ra": 226.0254,
            "Ac": 227.0278,
            "Th": 232.03806,
            "Pa": 231.03588,
            "U": 238.02891,
            "Np": 237.0482,
            "Pu": 244.0642,
            "Am": 243.0614,
            "Cm": 247.0703,
            "Bk": 247.0703,
            "Cf": 251.0796,
            "Es": 252.0829,
            "Fm": 257.0951,
            "Md": 258.0951,
            "No": 259.1009,
            "Lr": 262,
            "Rf": 267,
            "Db": 268,
            "Sg": 271,
            "Bh": 270,
            "Hs": 269,
            "Mt": 278,
            "Ds": 281,
            "Rg": 281,
            "Cn": 285,
            "Nh": 284,
            "Fl": 289,
            "Mc": 289,
            "Lv": 292,
            "Ts": 294,
            "Og": 294,
            "": 0,
        }

        return d[element]


    def _formula_parser(self, formula: str):
        """
        a chemical formular parser

        :param formula: formula string
        :returns: dictionary with element (key) and number (value)
        """
        s = findall("([A-Z][a-z]?)([0-9]*)", formula)
        d = dict(s)
        for k, v in d.items():
            if v == "":
                d[k] = 1
            else:
                d[k] = int(v)
        return d


    def molecular_weight(self, formula: str):
        """
        Calculate molecular weight using a chemical formula

        :param formula: chemical formula
        :returns: molecular weight

        ------
        Example
        ------
        >>> molecular_weight("C6H12O6")
        >>> 180.15588
        """
        "calculate molecular weight"
        d = self._formula_parser(formula)
        mw = sum([self.molar_mass(k) * v for k, v in d.items()])
        return mw


    def pure_unit(self, unit: str):
        """
        Get a pure unit string by removing elements in string (e.g., mmol C d-1)

        :param unit: unit string with elements
        :returns: pure unit string

        ------
        Example
        ------
        >>> pure_unit("mmol C d-1")
        >>> "mmol d-1"
        """
        # an element pool
        l = ["C", "N", "P", "O", "Fe", "Si", "Ca", "S"]
        # add leading whitespace
        l = [" " + i for i in l]

        # remove element in unit string
        for element in l:
            if element in unit:
                unit = unit.replace(element, "")

        return unit


    def format_base_unit(self, input_str, latex_format=True):
        """
        Convert the base unit to the desired format
        
        base unit means it is not a compound unit (e.g., m2, m-2, m/s, m s-1),
        which should be handled by the function `format_unit`
                
        :param input_str: a base unit string 
        :return: a formatted string
        
        ----------
        Example:
        
        >>> convert_a_unit('m2')
        '$m^2$'
        >>> convert_a_unit('m-2')
        '$m^{-2}$'
        """
        if input_str == 'n/a':
            return None

        ## if the input string contains a '^' character, then remove it
        if '^' in input_str:
            input_str = input_str.replace('^', '')

        if '**' in input_str:
            input_str = input_str.replace('**', '^')
            
        if 'o/oo' in input_str:
            input_str = input_str.replace('o/oo', '‰')
            
        if 'degree' in input_str:
            input_str = input_str.replace('degree', '°')

        ## when exponent is negative  (e.g., m-2, m-3)
        if '-' in input_str:
            parts = input_str.split('-')
            ## two parts: base and exponent
            base = parts[0]
            exponent = parts[1]
            
            if len(parts) == 2 and exponent.isdigit():
                exponent = int(exponent)
                if exponent != 0:
                    # Convert to the desired format using string formatting
                    # 3 layers of curly braces are needed to print a single curly brace
                    output_str = f"{parts[0]}$^{{-{exponent}}}$"
                    ## if not use LaTeX format, then remove dollar sign
                    if not latex_format: output_str = output_str.replace('$', '')                
                    return output_str
                
        ## when exponent is positive (e.g., m2, m3)
        ## also slice into two parts: base and exponent
        if len(input_str) > 1:
            base = input_str[:-1]
            exponent = input_str[-1]
            
            if exponent.isdigit():
                exponent = int(exponent)
                if exponent != 0:
                    # Convert to the desired format using string formatting
                    # 3 layers of curly braces are needed to print a single curly brace
                    output_str = f"{base}$^{{{exponent}}}$"
                    ## if not use LaTeX format, then remove dollar sign
                    if not latex_format: output_str = output_str.replace('$', '')
                    return output_str
        
        return input_str

    def format_unit(self, input, *args, **kwargs):
        """
        convert a unit in a string to the desired format, which can be used in plot labels and 
        facilitate the unit calculation
        
        :param input: a unit string 
        :return: a  formatted string
        
        ----------
        Example
        
        >>> format_unit('mol kg-1')
        'mol kg$^{-1}$'
        >>> format_unit('mmol m-2 d-1')
        'mmol m$^{-2}$ d$^{-1}$'
        """
        if input == 'n/a':
            return None
        unit_list = input.split(" ")
        unit_list = [self.format_base_unit(i, *args, **kwargs) for i in unit_list]

        clean_string = ' '.join(unit_list)
        return clean_string



# from pint import UnitRegistry
# import pathlib


# ureg = UnitRegistry()
# Q_ = ureg.Quantity
# file_path = pathlib.Path(__file__).parent.parent / "data/context.txt"
# ureg.load_definitions(file_path)



import pathlib

import numpy as np
import pandas as pd
import xarray as xr

from scipy.interpolate import (
    RegularGridInterpolator,
    LinearNDInterpolator,
    NearestNDInterpolator,
)
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


class GridOperation:

    
    def lon_n2g(self, x):
        """
        Change parts of observational latitude [100, 180] to GENIE longitude [-260, -180]
        Note it isn't axisymmetric!
        """
        # TBD: add value range checker

        if x > 100 and x < 180:
            return x - 360
        else:
            return x

    def lon_g2n(self, x):
        """
        Change parts of observational latitude [100, 180] to GENIE longitude [-260, -180]
        CANNOT simply use +80, or -80! It isn't axisymmetric!
        """
        # TBD: add value range checker

        if x < -180:
            return x + 360
        else:
            return x

    def normalise_obs_lon(self, data: xr.Dataset) -> xr.Dataset:
        return data.assign_coords(
            {"lon": list(map(self.lon_n2g, data.lon.values))}
        ).sortby("lon")

    def normalise_GENIE_lon(self, data: xr.Dataset) -> xr.Dataset:
        """
        Change parts of observational latitude [100, 180] to GENIE longitude [-260, -180]
        """
        ## if hasattr(data, "assign_coords"):
        return data.assign_coords(
            {"lon": list(map(self.lon_g2n, data.lon.values))}
        ).sortby("lon")

    def mask_Arctic_Med(self, array, policy="na"):
        """
        mask Arctic and Meditterean Sea in cGENIE modern continent configuration
        """
        if policy == "na":
            array[34:36, :] = np.nan
            array[27:30, 25:30] = np.nan
        elif policy == "zero":
            array[34:36, :] = 0
            array[27:30, 25:30] = 0

        return array

    def GENIE_grid_mask(
        self, base="worjh2", basin="ALL", subbasin="", mask_Arc_Med=False, invert=False
    ):
        """
        Get a modern GENIE 36x36 mask array from input data.
        The input array is flipped (left/right flip -> up/down flip) for easy recognition

        :continent: worjh2, worlg4, worbe2, GIteiiaa, GIteiiaa, p0055c
        :basin: Atlantic/Pacific/Indian/ALL/Tanzania
        :subbasin: N/S/ALL, ALL means Southern Ocean section included

        :returns: GENIE grid array where continent/ice cap is 0 and ocean is 1, default is 'worjh2'
        """

        file_path = (
            pathlib.Path(__file__).parent.parent
            / f"data/mask_{base}_{basin}{subbasin}.txt"
        )
        grid_mask_raw = np.loadtxt(file_path, dtype=int)
        grid_mask = np.flip(np.fliplr(grid_mask_raw))

        if mask_Arc_Med:
            grid_mask = mask_Arctic_Med(grid_mask, policy="zero")

        if invert:
            grid_mask = ~grid_mask + 2

        return grid_mask

    def normal_lon(self, N=36, edge=False):
        """
        Normal longitude in 10 degree resolution,
        if edge is False, then return midpoint
        """
        if edge:
            lon_edge = np.linspace(-180, 180, N + 1)
            return lon_edge
        else:
            lon = np.linspace(-175, 175, N)
            return lon

    def regrid_lat(self, x):

        """
        Transform <latitude> into cGENIE resolution to facilitate comparison between
        model and observational data.
        """
        if x >= -90 and x <= 90:
            lat_edge = np.rad2deg(np.arcsin(np.linspace(-1, 1, 37)))
            lat = np.rad2deg(np.arcsin(np.linspace(-1, 1, 36)))

            for i in range(36):
                if x > lat_edge[i] and x <= lat_edge[i + 1]:
                    x = lat[i]
        else:
            raise ValueError("Latitude must be in [-90,90]")

        return x

    def regrid_lon(self, x):

        """
        Transform <longitude> into cGENIE resolution to facilitate comparison between
        model and observational data.
        """

        if x >= -180 and x <= 180:
            lon_edge = np.linspace(-180, 180, 37)
            for i in range(36):
                if x > lon_edge[i] and x <= lon_edge[i + 1]:
                    x = (lon_edge[i] + lon_edge[i + 1]) / 2  # middle value in the bin
        else:
            raise ValueError("Longitude must be in [-180,180]")

        return x

    def haversine_distance(self, lat1, lon1, lat2, lon2):
        """
        Calculate the Haversine distance between two points.

        :param lat1: latitude of point 1
        :param lon1: longitude of point 1
        :param lat2: latitude of point 2
        :param lon2: longitude of point 2
        
        :returns: distance in km
        """
        # Convert latitude and longitude from degrees to radians
        lat1, lon1, lat2, lon2 = np.radians([lat1, lon1, lat2, lon2])

        # Haversine formula
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

        # Radius of the Earth in kilometers
        earth_radius = 6371.0
        distance = earth_radius * c

        return distance

    def geo_dis3d(self, point1, point2):
        """
        include the vertical distance

        :param point1: tuple/list of coordinates (z, lat, lon) or (lat, lon)
        :param point2: tuple/list of coordinates (z, lat, lon) or (lat, lon)
        :returns: distance in km
        """
        if len(point1) != 3 or len(point2) != 3:
            raise ValueError("Incompatible coordinates between two points")
        
        z1, lat1, lon1 = point1
        z2, lat2, lon2 = point2
        hor_dis = self.haversine_distance(lat1, lon1, lat2, lon2)
        ver_dis = abs(z1 - z2)/1000
        return np.sqrt(hor_dis**2 + ver_dis**2)

    def geo_dis2d(self, point1, point2):
        """
        include the vertical distance

        :param point1: tuple/list of coordinates (z, lon, lat) or (lon, lat)
        :param point2: tuple/list of coordinates (z, lon, lat) or (lon, lat)
        :returns: distance in km
        """
        if len(point1) != 2 or len(point2) != 2:
            raise ValueError("Incompatible coordinates between two points")
        
        lat1, lon1 = point1
        lat2, lon2 = point2
        return self.haversine_distance(lat1, lon1, lat2, lon2)


class Interporaltor:
    
    def __init__(self, dims, coordinates, values, grid_number=200, method="r-linear"):
        """
        initialize regridder

        :param array: xr data array or dataframe
        :param grid_number: number of target grid points
        :param method: interpolation method, only linear is supported
        """
        self.dims = dims
        self.coords = coordinates
        self.values = values

        ## number of grid points
        self.grid_number = grid_number
        ## create new coordinates
        self.gridded_coord = self.new_coordinate(n=grid_number)
        ## create meshgrid
        self.meshgrid = self.new_meshgrid(*self.gridded_coord, indexing="ij")
        ## interpolation function
        self.interp_function = self._create_interp_function(method=method)
        ## interpolate data
        self.gridded_data = self.interpolate_data(tuple(self.meshgrid))

    def _create_interp_function(self, method):
        """
        create interpolation function

        :param method: interpolation method, use "x-y" format. where x is strcutre or not, y is the algorithm
        """

        data_class = method.split("-")[0]
        ## regular grid interpolation
        if data_class == "r":
            true_method = method.split("-")[1]
            interp_function = RegularGridInterpolator(
                self.coords, self.values, method=true_method
            )
        ## irregular grid
        elif data_class == "ir":
            true_method = method.split("-")[1]
            if true_method == "linear":
                interp_function = LinearNDInterpolator(self.coords, self.values)
            elif true_method == "nearest":
                interp_function = NearestNDInterpolator(self.coords, self.values)
        else:
            raise ValueError("Method not supported")

        return interp_function

    def new_coordinate(self, n):
        "create new coordinates for regridding"
        new_coords = []
        for coord_values in self.coords:
            min_val, max_val = np.min(coord_values), np.max(coord_values)
            new_values = np.linspace(min_val, max_val, n)
            new_coords.append(new_values)
        return new_coords

    def new_meshgrid(self, *args, **kwargs):
        "regrid to finer dimensions"
        return np.meshgrid(*args, **kwargs)

    def interpolate_data(self, *args, **kwargs):
        "Use the interpolation function to get regridded values"
        return self.interp_function(*args, **kwargs)

    def to_xarray(self):
        """
        convert numpy array to xr data array
        """
        return xr.DataArray(
            self.gridded_data, dims=self.dims, coords=self.gridded_coord
        )



import numpy as np
import xarray as xr
from scipy.stats import sem
import regionmask

from .grid import Interporaltor, GridOperation
from .plot import ArrayVis

class GriddedData(ArrayVis):
    """
    GriddedData is a class to store and compute GENIE netcdf data.

    It stores data in xarray.DataArray format, and provides optimalised methods for GENIE model output to compute statistics.    
    """
    
    def __init__(self, arr=np.nan):
        """
        Initialise an instance of GriddedData

        Parameters
        ----------
        arr : xarray.DataArray
        """
        
        # set data
        self.array = arr
        
        # init plottable instance
        super().__init__(array=self.array)

    def __getitem__(self, item):
        "make GenieArray subscriptable like xarray.DataArray"
        return self.array[item]

    def interpolate(self, *args, **kwargs):
        """
        Interpolate the GENIE array to a finer grid, might be useful for plotting        
        """
        
        coords = tuple([self.array[dim].values for dim in self.array.dims])
        values = self.array.values
        dims = self.array.dims
        self.array = Interporaltor(dims, coords,values, *args, **kwargs).to_xarray()
        
        return self

    def sel(self, *args, **kwargs):
        """a wrapper to xarray `sel` method

        -----------
        Examples:
        -----------
        >>> model = GenieModel(path)
        >>> var = model.get_var(target_var)
        >>> var.search_grid(lon=XX, lat=XX, zt=XX, method='nearest')
        
        ## for loop over a data frame (df) with longitude, latitude, depth columns
        >>> modelvar = []
        >>> for i in range(len(df)):
        >>>     lat = df.latitude.iloc[i][0]
        >>>     lon = df.longitude[i].iloc[0]
        >>>     zt = df.depth[i].iloc[0]
        >>>     kwargs = {'lat': lat, 'lon':lon, 'zt': zt, 'method': 'nearest'}
        >>>     var = var.sel(**kwargs).values
        >>>     modelvar.append(var)        
        """
        
        self.array = self.array.sel(*args, **kwargs)
        return self


    def isel(self, *args, **kwargs):
        "a wrapper to xarray `isel` method"
        try:
            self.array = self.array.isel(*args, **kwargs)
            return self
        except:
            print("This array instance does not contain xarray.DataArray")
    
    def normalise_longitude(self):
        """
        Normalise GENIE's longitude (eastern degree) to normal longitude (-180, 180)

        -----------
        Example
        -----------

        >>> Model = GenieModel("a path")
        >>> Model.get_var('abc').normalise_longitude()
        """
        gp = GridOperation()
        self.array = gp.normalise_GENIE_lon(self.array)
        return self

    def __add__(self, other):
        """
        Allow GenieArray to be added by a number or another GenieArray
        """
        sum = GriddedData()
        if hasattr(other, "array"):
            sum.array = self.array + other.array
        else:
            ## a scalar
            sum.array = self.array + other
        return sum

    def __sub__(self, other):
        """
        Allow GenieArray to be subtracted by a number or another GenieArray
        """
        diff = GriddedData()
        if hasattr(other, "array"):
            diff.array = self.array - other.array
        else:
            ## a scalar
            diff.array = self.array - other
        return diff

    def __truediv__(self, other):
        """
        Allow GenieArray to be divided by a number or another GenieArray

        NA/NA -> NA
        """
        quotient_array = np.zeros_like(self.array)
        
        if isinstance(other, GriddedData):
            quotient_array = np.divide(self.array, other.array)
        else:
            # Handle division by a scalar
            quotient_array = np.divide(self.array, other)

        quotient = GriddedData()
        quotient.array = xr.DataArray(quotient_array)
        return quotient

    def __mul__(self, other):
        """
        Allow GenieArray to be multiplied by a number or another GenieArray
        """
        product = GriddedData()
        if hasattr(other, "array"):
            product.array = self.array * other.array
        else:
            try:
                product.array = self.array * other
            except ValueError:
                print("Sorry, only number and GenieArray are accepted")

        return product


    def max(self, *args, **kwargs):
        self.array = self.array.max(*args, **kwargs)
        return self


    def min(self, *args, **kwargs):
        self.array = self.array.min(*args, **kwargs)
        return self


    def sum(self, *args, **kwargs):
        ## restore units
        units = self.array.units
        self.array = self.array.sum(*args, **kwargs)
        self.array.units = units
        return self        


    def mean(self, *args, **kwargs):
        self.array = self.array.mean(*args, **kwargs)
        return self
    

    def median(self, *args, **kwargs):
        self.array = np.median(self.array, *args, **kwargs)
        return self        


    def sd(self, *args, **kwargs):
        "standard deviation of the mean"
        self.array = np.std(self.array, *args, **kwargs)
        return self        

    def variance(self, *args, **kwargs):
        self.array = np.var(self.array, *args, **kwargs)
        return self    

    def se(self, *args, **kwargs):
        "standard error of the mean"
        self.array = sem(self.array, nan_policy="omit", axis=None, *args, **kwargs)
        return self        
    
    def select_basin(self, basin):
        """
        select (modern) basin from regionmask.defined_regions.ar6.ocean

        The 58 defines marine regions from the sixth IPCC assessment report (AR6), Iturbide et al., (2020) ESSD
        
        https://regionmask.readthedocs.io/en/stable/_images/plotting_ar6_all.png
        
        --------------------------------
        47: North Pacific
        48: Equatorial Pacific
        49: South Pacific
        
        50: North Atlantic Ocean
        51: Equatorial Atlantic Ocean
        52: Southern Atlantic Ocean

        57: Southern Ocean
        --------------------------------        
        """
        ocean = regionmask.defined_regions.ar6.ocean
        index = ocean.map_keys(basin)
        mask = ocean.mask(self.normalise_longitude())
        self.array = self.normalise_longitude().array.where(mask == index)

        return self


    def mask_basin(self, base, basin, subbasin):

        """
        similar to select_basin, but use pre-defined grid mask
        """
        gp = GridOperation()
        data = self.array
        mask = gp.GENIE_grid_mask(base=base, basin=basin, subbasin=subbasin, invert=True)

        if self.array.ndim > 2:
            mask = np.broadcast_to(mask, (16, 36, 36))

        mask_data = np.ma.array(data, mask=mask)
        mask_data = np.ma.masked_invalid(mask_data)

        self.array = xr.DataArray(mask_data, dims=data.dims, coords=data.coords)        

        return self

    def ocn_only_data(self, index=False):
        """
        remove the NA grid (land in cGENIE definition)

        :param index: return GeoIndex or value
        """

        stacked_data = self.array.stack(x=self.array.dims)            
        stacked_ocn_mask =~np.isnan(stacked_data)
        ocn_only_data = stacked_data[stacked_ocn_mask]
        
        if index:
            return ocn_only_data.x.values
        else:            
            return ocn_only_data
        

    def search_grid(self, point, ignore_na=False, to_genielon=False):
        """
        search the nearest grid point to the given coordinates
        
        :param point: a list/tuple of coordinate values
        :param ignore_na: whether only check ocean data (which ignore the NA grids)
        :param to_genielon: whether convert the input longitude to genie longitude, input point must be list if True       
        """
        ## ignore the first dimension (which is time)        
        geo_ndim = self.array.ndim - 1
        
        if len(point) != geo_ndim:
            raise ValueError("Input point has incompatiable coordinate")        
        
        if ignore_na:
            
            match geo_ndim:
                case 3:
                    ## point: (zt, lat, lon); Index: (time, zt, lat, lon)
                    if to_genielon: point[2] = GridOperation().lon_g2n(point[2])
                    index_pool = self.ocn_only_data(index=True)                    
                    distances = np.array([GridOperation().geo_dis3d(point, pt[1:4]) for pt in index_pool])
                case 2:
                    ## point: (lat, lon); Index: (time, lat, lon)
                    if to_genielon: point[1] = GridOperation().lon_g2n(point[1])
                    index_pool = self.ocn_only_data(index=True)
                    distances = np.array([GridOperation().geo_dis2d(point, pt[1:3]) for pt in index_pool])

            idx_min = np.argmin(distances)
            nearest_value = self.ocn_only_data(index=False).values[idx_min]
            return nearest_value
        else:
            match geo_ndim:
                case 3:
                    zt, lat, lon = point
                    kwargs = {'lat': lat, 'lon':lon, 'zt': zt, 'method': 'nearest'}
                case 2:
                    lat, lon = point
                    kwargs = {'lat': lat, 'lon':lon, 'method': 'nearest'}
            return self.array.sel(**kwargs).values.item()

        
    def select_bbox(self, lon_min=-255, lon_max=95, lat_min=0, lat_max=90):
        """
        default longitude is unassigned of cGENIE grids
        """
        if lon_min > lon_max or lat_min > lat_max:
            raise ValueError("longitude/latitude min must be less than max!")

        lon = self.array.coords["lon"]
        lat = self.array.coords["lat"]

        return self.array.loc[
                dict(
                    lat=lat[(lat >= lat_min) & (lat <= lat_max)],
                    lon=lon[(lon >= lon_min) & (lon <= lon_max)],
                )
            ]
